#ifndef GTKLIFE_H
#define GTKLIFE_H

#include <gtk/gtk.h>

#include "util.h"

/* Defines and Enumerations */

#define PROG      "gtklife"
#define TITLE     "GtkLife"
#define USER_DIR  ".gtklife"
#define VERSION   "2.1"

#define INIT_X_CENTER             (WORLD_SIZE / 2)
#define INIT_Y_CENTER             (WORLD_SIZE / 2)
#define MIN_ZOOM                  1
#define MAX_ZOOM                  16
#define MAX_SKIPPED_FRAMES        10
#define MAX_RECENT_FILES          5
#define RECENT_FILES_INSERT_POS   9
#define DEFAULT_EXTENSION         "lif"
#define TABLE_OPTS                (GTK_SHRINK | GTK_FILL)
#define HELP_BROWSER_PARAM        "$f"

#define SCROLL_STEP  -1
#define SCROLL_PAGE  -2

enum color_index_enum  {BG_COLOR_INDEX, CELL_COLOR_INDEX, GRID_COLOR_INDEX, SELECT_COLOR_INDEX,
                        NUM_COLORS};

typedef enum tracking_mouse_enum {NOT_TRACKING, DRAWING, SELECTING, PASTING} tracking_mouse_type;

/* Configuration Defaults */

#define DEFAULT_PATTERNS_DIR      DATADIR "/patterns/"
#define DEFAULT_WINDOW_WIDTH      800
#define DEFAULT_WINDOW_HEIGHT     600
#define DEFAULT_ZOOM              1
#define DEFAULT_SHOW_TOOLBAR      TRUE
#define DEFAULT_SHOW_SIDEBAR      TRUE
#define DEFAULT_SHOW_SCROLLBARS   TRUE
#define DEFAULT_FULLSCREEN        FALSE
#define DEFAULT_BG_COLOR          0xC0C0C0
#define DEFAULT_CELL_COLOR        0x000080
#define DEFAULT_GRID_COLOR        0x808080
#define DEFAULT_SELECT_COLOR      0xFF0000
#define DEFAULT_SPEED             20
#define DEFAULT_SPEED_SLIDER_MIN  1
#define DEFAULT_SPEED_SLIDER_MAX  200
#define DEFAULT_SPEED_INCREMENT   10
#define DEFAULT_SKIP_FRAMES       TRUE
#define DEFAULT_HELP_BROWSER      "mozilla -remote 'openURL($f, new-window)' || mozilla '$f'"

/* Macros */

#define DRAW_GRID(zoom_level)  ((zoom_level) > 2)
#define SELECTION_ACTIVE()     (!rects_identical(&state.selection, &null_rect))

#define LEFT_ALIGN(w)   alignment = gtk_alignment_new(0, 0.5, 0, 0); \
                        gtk_container_add(GTK_CONTAINER(alignment), (w))
#define CENTER_ALIGN(w) alignment = gtk_alignment_new(0.5, 0.5, 0, 0); \
                        gtk_container_add(GTK_CONTAINER(alignment), (w))
#define FRAME_ALIGN(w)  alignment = gtk_alignment_new(0, 0, 1, 0); \
                        gtk_container_add(GTK_CONTAINER(alignment), (w))

/* Types */

typedef struct dimension_struct {
    int32 width;
    int32 height;
} dimension;

typedef struct point_struct {
    int32 x;
    int32 y;
} point;

typedef struct rect_struct {
    point start;
    point end;
} rect;

typedef struct recent_file_struct {
    char*       filename;
    char*       full_path;
    GtkWidget*  menu_item;
    GtkWidget*  label;
} recent_file;

typedef struct toolbar_item_struct {
    const char*    description;
    uint8*         icon;
    GtkSignalFunc  callback;
} toolbar_item;

typedef struct desc_dialog_info_struct {
    GtkWidget*  dialog;
    GtkWidget*  textbox;
} desc_dialog_info;

typedef struct speed_dialog_info_struct {
    GtkWidget*  dialog;
    GtkWidget*  speed_spinbox;
    GtkWidget*  skip_toggle;
} speed_dialog_info;

typedef struct color_dialog_info_struct {
    const char*  dialog_title;
    GtkWidget*   prefs_entry;
    uint32       cur_value;
    uint32       default_value;
} color_dialog_info;

typedef struct prefs_dialog_info_struct {
    GtkWidget*  dialog;
    GtkWidget*  patterns_dir_entry;
    GtkWidget*  default_window_width_spinbox;
    GtkWidget*  default_window_height_spinbox;
    GtkWidget*  default_zoom_combo;
    GtkWidget*  default_show_toolbar_toggle;
    GtkWidget*  default_show_sidebar_toggle;
    GtkWidget*  default_show_scrollbars_toggle;
    GtkWidget*  default_fullscreen_toggle;
    GtkWidget*  color_entries[NUM_COLORS];
    GtkWidget*  default_speed_spinbox;
    GtkWidget*  speed_slider_min_spinbox;
    GtkWidget*  speed_slider_max_spinbox;
    GtkWidget*  speed_increment_spinbox;
    GtkWidget*  default_skip_toggle;
    GtkWidget*  help_browser_entry;
    color_dialog_info*  color_dialog_infos[NUM_COLORS];
} prefs_dialog_info;

/* Prototypes */

/* Functions for reading/saving user settings */

void        create_user_dir(void);
void        load_preferences(void);
void        apply_defaults(void);
void        load_keybindings(void);
void        load_recent_files_list(void);
void        save_preferences(void);
void        save_keybindings(void);
void        save_recent_files_list(void);

/* Main loop functions */

void        main_loop(void);
void        tick_and_update(void);

/* GUI initialization functions */

void        init_gui(void);
void        init_rgb(void);
void        init_window(void);
void        init_menubar(GtkWidget* containing_box);
void        init_toolbar(GtkWidget* containing_box);
void        init_speed_box(GtkWidget* containing_box);
void        init_tick_display(GtkWidget* containing_box);
void        init_population_display(GtkWidget* containing_box);
void        init_panes(GtkWidget* containing_box);
void        init_sidebar(GtkWidget* containing_box);
void        init_canvas(GtkWidget* containing_box);

/* Misc. GUI functions */

void        adjust_scrollbars(void);
void        adjust_scrollbar_values(void);
void        error_dialog(const char* msg, ...);
void        full_canvas_redraw(void);
void        handle_mouse_press(GtkWidget *widget, GdkEventButton *event);
void        handle_mouse_release(GtkWidget *widget, GdkEventButton *event);
void        handle_mouse_movement(void);
void        handle_canvas_expose(GtkWidget *widget, GdkEventExpose *event);
void        handle_canvas_resize(void);
void        handle_hscrollbar_change(void);
void        handle_vscrollbar_change(void);
void        handle_sidebar_select(GtkCList *clist, gint row, gint column, GdkEventButton *event,
                                  gpointer user_data);
void        handle_speed_slider_change(void);
void        handle_zoom_change(void);
void        put_filename_in_window_title(void);
void        set_effective_canvas_size(void);
void        set_speed_label_size(int32 max_speed);
void        set_viewport_position(const point* pt, boolean center_around);
void        trigger_canvas_update(void);
void        update_population_label(void);
void        update_pattern_list(void);
void        update_tick_label(void);
GtkWidget*  create_spinbox(int32 init_value, int32 min, int32 max);

/* Functions for drawing the offscreen pixmap */

void        draw_life_pixmap(void);
void        draw_grid_and_boxes(void);
void        draw_screen_box(const rect* box);
void        draw_life_block(uint16 block, int32 xstart, int32 ystart, boolean full_update);
void        user_draw(const point* pt);
void        draw_from_to(const point* start, const point* end);

/* Bound Functions */

void        file_new(void);
void        file_open(void);
void        file_reopen(void);
void        file_save(void);
void        file_save_as(void);
void        file_description(void);
void        file_recent(GtkWidget* widget, int32* num);
void        file_quit(void);
void        view_recenter(void);
void        view_zoom_in(void);
void        view_zoom_out(void);
void        view_zoom_1(void);
void        view_zoom_2(void);
void        view_zoom_4(void);
void        view_zoom_8(void);
void        view_zoom_16(void);
void        view_fullscreen(void);
void        view_show_toolbar(void);
void        view_show_sidebar(void);
void        view_show_scrollbars(void);
void        edit_cut(void);
void        edit_copy(void);
void        edit_paste(void);
void        edit_move(void);
void        edit_clear(void);
void        edit_preferences(void);
void        run_start_stop(void);
void        run_step(void);
void        run_faster(void);
void        run_slower(void);
void        run_speed(void);
void        help_help(void);
void        help_about(void);

void        position_scroll_left(void);
void        position_scroll_right(void);
void        position_scroll_up(void);
void        position_scroll_down(void);
void        position_scroll_nw(void);
void        position_scroll_ne(void);
void        position_scroll_sw(void);
void        position_scroll_se(void);
void        position_page_left(void);
void        position_page_right(void);
void        position_page_up(void);
void        position_page_down(void);
void        position_page_nw(void);
void        position_page_ne(void);
void        position_page_sw(void);
void        position_page_se(void);

/* Bound Function Helpers */

void        position_scroll_generic(int32 xdir, int32 ydir, int32 how_far);
void        file_open_ok(GtkWidget* widget, GtkFileSelection* file_selector);
void        file_save_as_ok(GtkWidget* widget, GtkFileSelection* file_selector);
void        file_save_as_confirm_overwrite(GtkWidget* widget, GtkFileSelection* file_selector);
void        file_save_as_confirmed(GtkWidget* widget, GtkFileSelection* file_selector);
void        file_description_ok(GtkWidget* widget, desc_dialog_info* info);
boolean     file_description_apply(GtkWidget* widget, desc_dialog_info* info);
void        file_description_destroy(GtkWidget* widget, desc_dialog_info* info);
void        edit_paste_win_style(const point* pt);
void        edit_paste_unix_style(const point* pt);
boolean     edit_paste_verify_buffer(void);
boolean     edit_paste_verify_target(const point* pt);
void        edit_paste_perform(const point* pt);
void        edit_preferences_ok(GtkWidget* widget, prefs_dialog_info* info);
void        edit_preferences_apply(GtkWidget* widget, prefs_dialog_info* info);
void        edit_preferences_destroy(GtkWidget* widget, prefs_dialog_info* info);
void        run_speed_ok(GtkWidget* widget, speed_dialog_info* info);
void        run_speed_apply(GtkWidget* widget, speed_dialog_info* info);
void        run_speed_destroy(GtkWidget* widget, speed_dialog_info* info);
void        prefs_select_patterns_dir(GtkWidget* widget, prefs_dialog_info* info);
void        prefs_select_patterns_dir_ok(GtkWidget* widget, GtkFileSelection* file_selector);
void        prefs_patterns_dir(GtkWidget* widget, prefs_dialog_info* info);
void        prefs_select_color(GtkWidget* widget, color_dialog_info* info);
void        prefs_select_color_ok(GtkWidget* widget, GtkColorSelectionDialog* color_dialog);
void        prefs_default_color(GtkWidget* widget, color_dialog_info* info);
void        prefs_default_help_browser(GtkWidget* widget, prefs_dialog_info* info);

/* Miscellanous Functions */

void        add_to_recent_files(void);
boolean     attempt_load_pattern(const char* path);
boolean     attempt_save_pattern(const char* path);
void        get_logical_coords(const point* p, point* logical_p);
void        get_screen_coords(const point* p, point* screen_p);
void        get_screen_rectangle(const rect* r, rect* screen_r);
void        screen_box_update(rect* oldr, const rect* newr);
boolean     selection_check(void);
void        selection_copy_clear(boolean copy, boolean clear);
void        set_new_current_dir(const char* filepath);
void        set_speed(int32 new_speed);
void        sidebar_unselect(void);
void        start_stop(void);

/* Utility Functions */

void        alpha_blend(uint8* rgb1, const uint8* rgb2, int32 opacity);
void        get_rect_dimensions(const rect* r, dimension* dim);
uint64      get_time_milliseconds(void);
boolean     is_numeric(const char* str);
GtkWidget*  load_pixmap_from_xpm_file(const char* path);
GtkWidget*  load_pixmap_from_rgba_array(uint8* pixels, int32 width, int32 height, GdkColor* bg);
void        normalize_rectangle(rect* r);
boolean     points_identical(const point* p1, const point* p2);
boolean     rects_identical(const rect* r1, const rect* r2);

#endif
